
// The X11 graphics module:
//
// This module has been greatly enhanced by the content of "Xlib by Example",
// by Cui-Qinq Yang and Mahir S. Ali.


#include <unistd.h>
#include <stdio.h>

#include "x11_framebuf.h"
#include "cursor.h"

// Not in any header file...
extern "C" {
	extern int XShmQueryExtension(Display *dpy);
}


/* Shared memory error handler routine */
static int haderror;
static int (*origerrorhandler)(Display *, XErrorEvent *);
static int shmerrorhandler(Display *d, XErrorEvent *e)
{
	haderror++;
	if(e->error_code==BadAccess)
		fprintf(stderr,"X: failed to attach shared memory\n");
	else (*origerrorhandler)(d,e);
} 

X11_FrameBuf:: X11_FrameBuf(char **icon, 
		unsigned short width, unsigned short height, char *title) :
					FrameBuf(icon, width, height, title)
{
	XSetWindowAttributes winattr;
	XVisualInfo     vinfo_return;
	XGCValues       gcv;
	char           *window_Name = title;
        XTextProperty   windowName;
	XWMHints       *wmhints;
        XClassHint     *classhints;
        XSizeHints     *sizehints;
	XpmAttributes   attributes;
	Pixmap		icon_pixmap;
	Pixmap		cursor_pixmap;
	Pixmap		cursor_mask;
	Pixmap		nocursor_pixmap;
	Pixmap		nocursor_mask;
	XColor		black, white;
        unsigned long   valuemask=0L;
	int		XpmErrorStatus;
	XEvent          event;
	char           *argv[2] = { title, NULL };

	/* Open our DISPLAY */
	if ( (display=XOpenDisplay(NULL)) == NULL ) {
		fprintf(stderr, "X: Can't open display...\n");
		exit(3);
	}

	/* Make sure all is destroyed if killed off */
	XSetCloseDownMode(display, DestroyAll);

	/* Make sure we can do PsuedoColor colormap stuff */
	if ( XMatchVisualInfo(display, DefaultScreen(display),
			8, PseudoColor, &vinfo_return) == False ) {
		fprintf(stderr, "X: Screen doesn't support PseudoColor!\n");
		exit(3);
	}
	colormap = DefaultColormap(display, DefaultScreen(display));

	/* Check to see if the extensions are supported */
	if (!XShmQueryExtension(display)) {
		fprintf(stderr, "X server doesn't support MITSHM extension.\n");
		exit(3);
	}
	if (!XShmPixmapFormat(display)) {
		fprintf(stderr, "X server doesn't do shared memory pixmaps.\n");
		exit(3);
	}

	/* Create the main window */
	Black = BlackPixel(display, DefaultScreen(display));
	MainWin = XCreateSimpleWindow(display, DefaultRootWindow(display),
				0, 0, width, height, 1, Black, Black);

	/* Get our shared screen memory */
	shared_len = (width * height);	// 8 bits per pixel
	backbuf = new unsigned char[shared_len];
	shminfo.shmid = shmget(IPC_PRIVATE, shared_len, IPC_CREAT|0777);
	if ( shminfo.shmid < 0 ) {
		fprintf(stderr, "Couldn't get X shared memory.\n");
		exit(3);
	}
/* We need to protect our shared memory against orphanage if we
   are killed in this function.  */

	shminfo.shmaddr = (char *) shmat(shminfo.shmid, 0, 0);
	shared_mem = (unsigned char *) shminfo.shmaddr;
	if ( shminfo.shmaddr == (char *)-1 ) {
		fprintf(stderr, "Couldn't attach to X shared memory.\n");
		shmctl(shminfo.shmid, IPC_RMID, 0);
		exit(3);
	}
	shminfo.readOnly = False;

	/* Now try to attach it to the X Server */
	haderror = False;
	origerrorhandler = XSetErrorHandler(shmerrorhandler);
	XShmAttach (display, &shminfo);
	XSync(display,True);  /* wait for error or ok */
	XSetErrorHandler(origerrorhandler);
	if( haderror ){
		if ( shmdt(shminfo.shmaddr) < 0 )
			perror("X shmdt() error");
		if ( shmctl(shminfo.shmid, IPC_RMID, 0) < 0 )
			perror("X shmctl rmid() error");
		exit(3);
	}
	memset((void *)backbuf, Black, shared_len);
	memset((void *)shared_mem, Black, shared_len);

	/* Create the shared memory Pixmap */
	X_pixmap = XShmCreatePixmap(display, MainWin, shminfo.shmaddr,
						&shminfo, width, height, 8);
	XSetWindowBackgroundPixmap(display, MainWin, X_pixmap);

	/* Create the icon window, and give it the icon colormap */
	wmhints = XAllocWMHints();
	if ( icon ) {
		/* Allocate the pixmap for the icon */
		attributes.valuemask = XpmColormap;
		attributes.colormap = colormap;
        	XpmErrorStatus = XpmCreatePixmapFromData(display, MainWin, 
					icon, &icon_pixmap, NULL, &attributes);
		/* Ignore a failed pixmap creation */
		if ( XpmErrorStatus != XpmSuccess ) {
			fprintf(stderr, 
				"Warning: Couldn't create color icon.\n");
		}
		XpmFreeAttributes(&attributes);

		wmhints->icon_pixmap 	= icon_pixmap;
		wmhints->flags = IconPixmapHint;
	} else
		wmhints->flags = 0;

	/* Create the normal cursor (Mac cursor) */
	cursor_pixmap = XCreatePixmapFromBitmapData(display, MainWin, 
			cursor_bits, cursor_width, cursor_height, 1L, 0L, 1);
	cursor_mask = XCreatePixmapFromBitmapData(display, MainWin, 
			cursorm_bits, cursorm_width, cursorm_height, 1L, 0L, 1);
	black.red   = 0x0000;
	black.green = 0x0000;
	black.blue  = 0x0000;
	white.red   = 0xFFFF;
	white.green = 0xFFFF;
	white.blue  = 0xFFFF;
	cursor  = XCreatePixmapCursor(display, cursor_pixmap,
					cursor_mask, &black, &white, 0, 0);
	XFreePixmap(display, cursor_pixmap);
	XFreePixmap(display, cursor_mask);

	/* Create the hidden cursor */
	black.red   = 0;
	black.green = 0;
	black.blue  = 0;
	attributes.valuemask = (XpmColormap|XpmDepth);
	attributes.colormap = colormap;
	attributes.depth = 1;
	XpmCreatePixmapFromData(display, MainWin, invis_cursor_xpm, 
				&nocursor_pixmap, &nocursor_mask, &attributes);
	XpmFreeAttributes(&attributes);
	nocursor  = XCreatePixmapCursor(display, nocursor_pixmap,
					nocursor_mask, &black, &black, 0, 0);
	XFreePixmap(display, nocursor_pixmap);
	XFreePixmap(display, nocursor_mask);

	/* Set the cursor */
	hidden_cursor = 1;
	Show_Cursor();

	/* Various window manager settings */
        wmhints->initial_state   = NormalState;
        wmhints->input           = True;     
        wmhints->flags |= StateHint | InputHint;

	/* Set the class for this program */
	classhints = XAllocClassHint();
	classhints->res_name     = title;
	classhints->res_class    = title;

        /* Setup the max and minimum size that the window will be */
	sizehints = XAllocSizeHints();
        sizehints->flags         = PSize | PMinSize | PMaxSize;
        sizehints->min_width     = width;
        sizehints->min_height    = height;
        sizehints->max_width     = width;
        sizehints->max_height    = height;
	WIDTH = width;
	HEIGHT = height;

	/* Create the window/icon name properties */
        if (XStringListToTextProperty(&window_Name, 1, &windowName) == 0) {
		fprintf(stderr, "X: Cannot create window name resource!\n");
		exit(3);
	}

        /* Now set the window manager properties */
        XSetWMProperties(display, MainWin, &windowName, &windowName,
				argv, 1, sizehints, wmhints, classhints);
	XFree(wmhints);
	XFree(classhints);
	XFree(sizehints);

  	/*  Enable the delete window protocol (taken from 'rxvt' -- Thanks!) */
	wm_del_win = XInternAtom(display,"WM_DELETE_WINDOW",False);
	XSetWMProtocols(display, MainWin, &wm_del_win, 1);
  
        /* Check if the server allows backing store */
        if (DoesBackingStore(XDefaultScreenOfDisplay(display)) == Always)
        {
                /* Ok we want backing store as it is very useful */
                valuemask |= CWBackingStore;
                winattr.backing_store = Always;
        }
	XChangeWindowAttributes(display, MainWin, valuemask, &winattr);

	/* Set up graphics contexts */
	gcv.graphics_exposures = False;
        if (!(gc = XCreateGC(display, MainWin, GCGraphicsExposures, &gcv))) {
		fprintf(stderr, "X: Couldn't create graphics context.\n");
		exit(3);
	}

	/* Set up the events we wait for */
	Event_mask =    KeyPressMask | KeyReleaseMask |
                        ButtonPressMask | ButtonReleaseMask |
                        // PointerMotionMask |
                        ExposureMask | StructureNotifyMask;
	XSelectInput(display, MainWin, Event_mask);

	/* Actually map the main window */
	XMapWindow(display, MainWin);

	/* Loop, waiting for our window to be mapped */
	do {
		/* Get the next event */
		XNextEvent(display, &event);
	}
	while (event.type != MapNotify);
}

X11_FrameBuf:: ~X11_FrameBuf()
{
	XUnmapWindow(display,  MainWin);
	XFreePixmap(display, X_pixmap);
	XFreeCursor(display, cursor);
	XFreeCursor(display, nocursor);
	XSync(display, False);
	XShmDetach (display, &shminfo);
	XSync(display, False);	/* need server to detach so can remove id */
	if ( shmdt(shminfo.shmaddr) < 0 )
		perror("X shmdt() error");
	if ( shmctl(shminfo.shmid, IPC_RMID, 0) < 0 )
		perror("X shmctl(rmid) error");
	XCloseDisplay(display);
}

/* Just try to allocate a full colormap */
/* This code assumes a psuedo-color (256 color) display */
int
X11_FrameBuf:: Alloc_Cmap(int polite, Color Cmap[NUM_COLORS])
{
	int i, c, r, g, b;
	int alloct[NUM_COLORS], nalloct;
	unsigned long pix;
	XColor xcol, cmap[NUM_COLORS];

	/* Most of the mapping code is adapted from 'xv' -- Thanks! :) */
	int ri, gi, bi;			/* The color shades we want */
	int rd, gd, bd;			/* The color shades we have */
	int mdist, close, d;		/* Distance and closest pixel */

	/* Shed light on a standard colormap, if we can */
	for ( i=0; i<NUM_COLORS; ++i )
		alloct[i] = 0;

	for ( r=0; r<6; ++r ) {
		for ( g=0; g<6; ++g ) {
			for ( b=0; b<6; ++b ) {
				xcol.red =   (((r*255)/5)<<8);
				xcol.green = (((g*255)/5)<<8);
				xcol.blue =  (((b*255)/5)<<8);
				xcol.flags = (DoRed|DoGreen|DoBlue);
				if ( XAllocColor(display, colormap, &xcol) ) {
					alloct[xcol.pixel]=1;
					++nalloct;
				}
			}
		}
	}
#ifdef DEBUG
fprintf(stderr, "Phase 1: nalloct = %d\n", nalloct);
#endif

	/* If we couldn't allocate an orthogonal plane of color */
	if ( (nalloct < (6*6*6)) && (nalloct > 128) ) {
		/* Free allocated colors and allocate a more limited range */
		for ( i=0; i<NUM_COLORS; ++i ) {
			if ( alloct[i] ) {
				pix = i;
				XFreeColors(display, colormap, &pix, 1, 0L);
				alloct[i]=0;
				--nalloct;
			}
		}
		for ( r=0; r<4; ++r ) {
			for ( g=0; g<8; ++g ) {
				for ( b=0; b<4; ++b ) {
					xcol.red =   (((r*255)/3)<<8);
					xcol.green = (((g*255)/7)<<8);
					xcol.blue =  (((b*255)/3)<<8);
					xcol.flags = (DoRed|DoGreen|DoBlue);
					if ( XAllocColor(display,
							colormap, &xcol) ) {
						alloct[xcol.pixel]=1;
						++nalloct;
					}
				}
			}
		}
	}
#ifdef DEBUG
fprintf(stderr, "Phase 2: nalloct = %d\n", nalloct);
#endif

	/* Find out what kind of colormap we have now */
	for ( i=0; i<NUM_COLORS; ++i ) {
		cmap[i].pixel = i;
		cmap[i].flags = (DoRed|DoGreen|DoBlue);
	}
	XQueryColors(display, colormap, cmap, NUM_COLORS);

	/* Map the whole colormap */
	for ( i=0; i<NUM_COLORS; ++i ) {
		mdist = 100000; close=0;

		ri = (Cmap[i].red >> 8);
		gi = (Cmap[i].green >> 8);
		bi = (Cmap[i].blue >> 8);

		/* Cycle through the colors we have */
		for ( c=0; c<NUM_COLORS; ++c ) {
			rd = ri - (cmap[c].red >> 8);
			gd = gi - (cmap[c].green >> 8);
			bd = bi - (cmap[c].blue >> 8);
			d = rd*rd + gd*gd + bd*bd;

			if ( d < mdist ) {
				mdist = d;
				close = c;
			}
		}
#ifdef DEBUG
fprintf(stderr, "Desired RGB (%d,%d,%d) mapped to existing RGB (%d,%d,%d)\n",
				ri, gi, bi, cmap[close].red>>8,
				cmap[close].green>>8, cmap[close].blue>>8);
#endif
		/* Lock down and register the color we want */
		if ( ! XAllocColor(display, colormap, &cmap[close]) ) {
			fprintf(stderr,
	"X11_FrameBuf: Warning: couldn't allocate an existing color!\n");
		}
		Pixel_colors[i] = cmap[close].pixel;
		Color_Map[i].red   = cmap[close].red;
		Color_Map[i].green = cmap[close].green;
		Color_Map[i].blue  = cmap[close].blue;
	}
	return(NUM_COLORS);
}

/* Allocate a private colormap :) */
int
X11_FrameBuf:: Alloc_Private_Cmap(Color Cmap[NUM_COLORS])
{
	int i;
        XColor xcols[NUM_COLORS];

	/* Create a custom visual colormap, if needed */
	private_cmap = XCreateColormap(display, MainWin,
			DefaultVisual(display, XDefaultScreen(display)),
							AllocAll);

	/* Allocate custom colors... */
	for(i=0;i<NUM_COLORS;i++)
		xcols[i].pixel = i;
	/*XQueryColors(display, private_cmap, xcols, NUM_COLORS);*/
	for(i=0;i<NUM_COLORS;i++) {
		if ( !Cmap[i].red && !Cmap[i].green && !Cmap[i].blue )
			Black = i;
		Pixel_colors[i] = i;
		Color_Map[i].red = xcols[i].red   = Cmap[i].red<<8;
		Color_Map[i].green = xcols[i].green = Cmap[i].green<<8;
		Color_Map[i].blue = xcols[i].blue  = Cmap[i].blue<<8;
		xcols[i].flags = (DoRed|DoGreen|DoBlue);
	}
	memset((void *)backbuf, Black, shared_len);
	XStoreColors(display, private_cmap, xcols, NUM_COLORS);
	XSetWindowColormap(display, MainWin, private_cmap);
	return(NUM_COLORS);
}

void
X11_FrameBuf:: Hide_Cursor(void)
{
	if ( hidden_cursor ==  0 ) {
		XDefineCursor(display, MainWin, nocursor);
		hidden_cursor = 1;
	}
}

void
X11_FrameBuf:: Show_Cursor(void)
{
	if ( hidden_cursor == 1 ) {
		XDefineCursor(display, MainWin, cursor);
		hidden_cursor = 0;
	}
}

/* This function was adapted from 'xscreensaver', by Jamie Zawinski 

	-- Thanks! :)
*/
void
X11_FrameBuf:: Fade(int steps)
{
	static int      state = XFADE_IN;
	static XColor   orig_colors[NUM_COLORS];
	XColor          curr_colors[NUM_COLORS];
	int             i, j;
	static Colormap *fade_cmap, new_cmap;

	/* Set the requested pixels */
	for ( i=0; i<NUM_COLORS; ++i )
		orig_colors[i].pixel = i;

	state = ( (state == XFADE_IN) ? XFADE_OUT : XFADE_IN );
	/* Make a copy of the default colormap and use that in fade */
	if ( private_cmap ) {
		if ( state == XFADE_OUT ) {
			XQueryColors(display, private_cmap, orig_colors, 
								NUM_COLORS);
			fade_cmap = &private_cmap;
		}
	} else {
		XQueryColors(display, colormap, orig_colors, NUM_COLORS);
		if ( state == XFADE_OUT ) {
			new_cmap = XCreateColormap(display,
				DefaultRootWindow(display),
				DefaultVisual(display, XDefaultScreen(display)),
								AllocAll);
			XStoreColors(display,new_cmap,orig_colors,NUM_COLORS);
			XGrabServer(display);
			XInstallColormap(display, new_cmap);
			fade_cmap = &new_cmap;
		}
	}
	XSync(display, False);

	memcpy(curr_colors, orig_colors, NUM_COLORS*sizeof(XColor));
	if ( state == XFADE_OUT ) {
		for ( i=steps; i > 0; i-- ) {
			for (j = 0; j < NUM_COLORS; j++) {
				curr_colors[j].red   = 
					orig_colors[j].red   * i / steps;
				curr_colors[j].green = 
					orig_colors[j].green * i / steps;
				curr_colors[j].blue  = 
					orig_colors[j].blue  * i / steps;
			}
			XStoreColors (display, *fade_cmap, 
						curr_colors, NUM_COLORS);
			usleep(10);
			XSync (display, False);
		}
	} else {
		for ( i=0; i < steps; i++ ) {
			for (j = 0; j < NUM_COLORS; j++) {
				curr_colors[j].red   = 
					orig_colors[j].red   * i / steps;
				curr_colors[j].green = 
					orig_colors[j].green * i / steps;
				curr_colors[j].blue  = 
					orig_colors[j].blue  * i / steps;
			}
			XStoreColors (display, *fade_cmap, 
						curr_colors, NUM_COLORS);
			usleep(10);
			XSync (display, False);
		}
	}
	
	if ( state == XFADE_IN ) {
		/* Restore the original colormap */
		if ( private_cmap ) {
			XStoreColors (display, private_cmap, 
						orig_colors, NUM_COLORS);
		} else {
			XInstallColormap(display, colormap);
			XUngrabServer(display);
			XFreeColormap(display, *fade_cmap);
		}
		XSync(display, False);
	}
}
	
void
X11_FrameBuf:: RefreshArea(int x0, int y0, int width, int height)
{
	XClearArea(display, MainWin, x0, y0, width, height, False);
}

void
X11_FrameBuf:: Refresh(void)
{
	XClearWindow(display, MainWin);
}

void
X11_FrameBuf:: FlushEvents(void)
{
	int    numevents;
	XEvent event;

	for ( numevents=XPending(display); numevents > 0; --numevents )
		GetEvent(&event);
}
	
int
X11_FrameBuf:: NumEvents(void)
{
	return(XPending(display));
}

void
X11_FrameBuf:: GetEvent(XEvent *event)
{
	XNextEvent(display, event);

	/* Handle some events internally */
	switch (event->type) {
		/* Are we being sent a message? */
		case ClientMessage:
			if ( event->xclient.format == 32 && 
					event->xclient.data.l[0] == wm_del_win )
			{ /* Window manager wants to delete us */
			  /* Assume that atexit() handles all cleanup */
				exit(0);
			}
			break;

		/* Cover yourself! :) */
		case Expose:
			Refresh();
			break;
	}
}

int
X11_FrameBuf:: KeyToAscii(XEvent *event, char *buf, int buflen, KeySym *key)
{
	return(XLookupString(&event->xkey, buf, buflen, key, NULL));
}

void
X11_FrameBuf:: Flush(int sync)
{
	if ( sync )
		XSync(display, False);
	else
		XFlush(display);
}
