
/* X11 Graphics routines for Maelstrom!  (By Sam Lantinga) */

#ifndef _x11_framebuf_h
#define _x11_framebuf_h

#include "framebuf.h"

#include <X11/Xutil.h>
#include <X11/xpm.h>
#include <sys/ipc.h>
#include <sys/shm.h>
#include <X11/extensions/XShm.h>


#define XFADE_OUT	0
#define XFADE_IN	1

/* This class allocates a 640x480 frame-buffer and provides high-level
   routines to blit images into it.  Well, it also provides access to
   other X11 based drawing routines.
*/
class X11_FrameBuf : public FrameBuf {

public:
	X11_FrameBuf(char **icon,
		unsigned short width, unsigned short height, char *title);
	~X11_FrameBuf();

	int            Alloc_Cmap(int polite, Color Cmap[]);
	int            Alloc_Private_Cmap(Color Cmap[]);
	void           RefreshArea(int x0, int y0, int width, int height);
	void           Refresh(void);
	void           Hide_Cursor(void);
	void           Show_Cursor(void);
	void           Fade(int steps);
	void           FlushEvents(void);
	int            NumEvents(void);
	void           GetEvent(XEvent *event);
	int            KeyToAscii(XEvent *event, char *buf, int buflen,
								KeySym *key);
	void           Flush(int sync);

protected:
	Display *display;		// The X11 display connection
	Window   MainWin;		// Our window on the display
	unsigned long Event_mask;	// The events we wait for
	XImage  *xwindow;		// Our in-core image of our window
	Colormap colormap;		// The current colormap
	Colormap private_cmap;		// A private colormap?
	Atom     wm_del_win;		// The delete window flag
	GC       gc;			// The current graphics context

	XShmSegmentInfo shminfo;	// The shared memory segment info
	Pixmap X_pixmap;		// The shared background pixmap

	Cursor cursor, nocursor;
};
#endif /* _x11_framebuf_h */
